#!/usr/bin/perl -w

use strict;
use warnings;

use File::Basename;
use Pod::Usage;
use Getopt::Long;
use Config::Tiny;
use Net::SSH qw(ssh_cmd);
use Sys::Hostname::Long;
use Socket;
use IO::File;

my $ME = basename $0;

my $opt_rules_file = '/etc/send-config/filter.rules';
my $opt_config     = '/etc/send-config/config';
my $opt_dry_run    = 0;
my $opt_keepgoing  = 0;
my $opt_verbose    = 0;
my $opt_debug      = 0;

sub debug;
sub readConfig($);
sub rsync;
sub readFilter($);

MAIN: {

    GetOptions(
        "r|rules-file=s" => \$opt_rules_file,
        "c|config=s"     => \$opt_config,
        "k|keepgoing!"   => \$opt_keepgoing,
        "n|dry-run!"     => \$opt_dry_run,
        "v|verbose!"     => \$opt_verbose,
        "d|debug!"       => \$opt_debug,
        "h|help"         => sub { pod2usage( -verbose => 0, -exitval => 0 ) }
    ) or pod2usage();

    unless ( scalar(@ARGV) == 1 ) {
        pod2usage( -verbose => 0, -exitval => 0 );
    }

    my $dest    = $ARGV[0];
    my $dest_ip = scalar gethostbyname($dest);

    ($dest_ip) or die "$ME: Can't resolve $dest\n";

    my $source_host = hostname_long();

    debug(  "PROG destination host: [" 
          . $dest . "]" 
          . " ip: ["
          . inet_ntoa($dest_ip)
          . "]" )
      if $opt_debug;
    debug( "PROG source host: [" . $source_host . "]" ) if $opt_debug;

    my ( $username, $path, $stamp ) = readConfig($opt_config);

    # adding trailing / to destination directory
    $path .= "/" unless ( $path =~ /\/$/ );
    my $dest_path = $path . $source_host;

    # create remote destination directory
    my $ssh_cmd_args = {
        user    => $username,
        host    => $dest,
        command => 'mkdir',
        args    => [ '-m 0700', "-p $dest_path" ],
    };
    ssh_cmd $ssh_cmd_args unless $opt_dry_run;

    debug( "PROG ssh command: ssh $username\@$dest $ssh_cmd_args->{command} ",
        join ' ', @{ $ssh_cmd_args->{args} } )
      if $opt_debug;

    rsync( $username, $dest, $dest_path, $stamp );

}

sub debug {
    print map( "DEBUG: $_\n", @_ );
}

sub readConfig($) {

    my $file = shift;

    my $config = Config::Tiny->read($file) or die "$ME: Can't open $file: $!\n";

    my $username = $config->{_}->{username} ? $config->{_}->{username} : 'root';
    my $path =
      $config->{_}->{path} ? $config->{_}->{path} : '/root/Configs/Hosts/';
    my $stamp =
        $config->{_}->{stamp}
      ? $config->{_}->{stamp}
      : '/var/lib/get-config.stamp';

    debug(
        "CONF remote username: $username",
        "CONF destination path: $path",
        "CONF stamp path: $stamp"
    ) if $opt_debug;

    return ( $username, $path, $stamp );

}

sub rsync() {

    my ( $username, $dest, $dest_path, $stamp ) = @_;
    my @status = ();

    my @cmd = (
        qw(rsync --rsh), "ssh -x",
        qw(--compress --numeric-ids
          --delete --delete-excluded
          --archive --relative)
    );

    push @cmd, "--verbose" if $opt_verbose;
    push @cmd, "--dry-run" if $opt_dry_run;

    my @filter = readFilter($opt_rules_file);

    my @src = map { /^\+\s*(.*)/ ? ( glob($1) ) : () } @filter;

    if (
        my @excludes =
        map { /^-\s*(.*)/ ? ( "--exclude" => glob($1) ) : () } @filter
      )
    {
        push @cmd, @excludes;
    }

    push @cmd, @src, "$username\@$dest:$dest_path";

    debug( "PROG rsync command: " . join( " ", @cmd ) ) if $opt_debug;

    open( TMP, "+>/tmp/$ME.$$" )
      or die "$ME: Can't open /tmp/$ME.$$";

    my $pid = fork();
    die "Can't fork" if not defined $pid;

    if ( $pid == 0 ) {
        open( STDERR, ">&TMP" ) or die "$!";
        open( STDOUT, ">/dev/null" ) unless $opt_verbose;
        exec @cmd;
        warn "Failed to exec '", join ' ', @cmd, "': $!\n";
    }
    else {
        waitpid $pid, 0;
    }

    close(TMP);

    if ($?) {

        open( TMP, "+</tmp/$ME.$$" )
          or die "$ME: Can't open /tmp/$ME.$$";
        unlink "/tmp/$ME.$$";

        while (<TMP>) {
            chomp;
            push @status, "[WARNING] $_"
              unless $opt_keepgoing and !$opt_dry_run;
        }
        close(TMP);
    }

    if (@status) {
        unless ($opt_keepgoing) {
            warn join( "\n", @status ), "\n";
            warn "[WARNING] $stamp not touched!\n";
            warn "\nUse option --keepgoing, to ignore this warning.\n";
        }
    }
    else {
        open( STAMP, ">$stamp" )
          or die "$ME: Can't open: $stamp: $!\n"
          unless $opt_dry_run;
        close(STAMP) unless $opt_dry_run;
        warn "[OK] touched $stamp.\n" unless $opt_dry_run;
    }

}

sub readFilter($) {

    my $file = shift;
    my @filter;

    my $cf = new IO::File $file or die "$ME: Can't open $file: $!\n";

    while (<$cf>) {
        next if /^(;#\s*$)/;
        push @filter, $_;
    }

    return @filter;

}

__END__

=pod

=head1 NAME

B<send-config> - copies files and directories to a remote machine

=head1 SYNOPSIS

B<send-config> [OPTION...] DEST

=over

=item B<-r, --rules-file=FILE>

=item B<-c, --config=FILE>

=item B<-n, --dry-run>

=item B<-k, --keepgoing>

=item B<-v, --verbose>

=item B<-d, --debug>

=item B<-h, --help>

=back

=head1 DESCRIPTION

B<send-config> copies files and directories to a remote machine. It uses L<rsync(1)> for data transfer
and uses the same authentication and provides the same security as L<ssh(1)>.

=over

=item B<DEST>

remote machine

=back

=head1 OPTIONS

=over 7

=item B<-r, --rules-file=FILE>

read exclude/include patterns from FILE (default: F</etc/send-config/filter.rules>)

=item B<-c, --config=FILE>

configuration file for send-config (default: F</etc/send-config/config>)

=item B<-n, --dry-run>

perform a trial run with no changes made

=item B<-k, --keepgoing>

don't stop on warnings

=item B<-v, --verbose>

explain what is being done

=item B<-d, --debug>

print debug informations

=item B<-h, --help>

display short help and exit

=back

=head1 EXAMPLES

F</etc/send-config/filter.rules>

Lines started with an # or ; are comments. New lines will be ignored. Lines with prefix + or - will be (+ included) or (- excluded) by the rsync process.

 - *~
 - /usr/src
 - /etc/samba/codepages
 - /usr/local/uvscan
 - /usr/local/src
 - /root/local-packages

 + /etc
 + /var/cache/debconf
 + /var/lib/dpkg/status
 + /boot/config*
 + /boot/grub/menu.lst
 + /root/.bash*
 + /root/.profile
 + /root/LOG*
 + /root/.ssh
 + /proc/config.gz
 + /usr/local
 + /usr/lib/AntiVir/*.[Kk][Ee][Yy]
 + /usr/src/*/*.config
 + /usr/src/*config*

F</etc/send-config/config>

I<username> is the remote login user for ssh and rsync process. I<path> is the absolute destination path on the remote machine where the files will be stored. I<stamp> is the absolute path on the local machine to create or touch an stamp file, you can use it e.g. for nagios (check_file_age).

 username = root
 path = /root/Configs/Hosts/
 stamp = /var/lib/get-config.stamp

=head1 FILES

F</etc/send-config/config>

F</etc/send-config/filter.rules>

=head1 AUTHOR

Christian Arnold <arnold@schlittermann.de>

=head1 COPYRIGHT

Copyright (C) 2010 Schlittermann - internet & unix support. License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>

This is free software: you are free to change and redistribute it.  There is NO WARRANTY, to the extent permitted by law.

=head1 SEE ALSO

L<ssh(1)>, L<rsync(1)>

=cut

